-- Modified 12/6/98 by tmoran@bix.com to use CLAW instead of ANSI.SYS

--::::::::::
--room.adb
--::::::::::
with Ada.Numerics.Elementary_Functions,
     Claw.Basic_Window,
     Claw.Brushes,
     Claw.Canvas,
     Claw.Draw,
     Claw.Icons,
     Claw.Pens,
     Claw.Window_Operations;
with Chop;
with Phil;
with Society;
pragma Elaborate (Phil);
package body Room is

  -- Dining Philosophers, Ada 95 edition
  -- A line-oriented version of the Room package
  -- Michael B. Feldman, The George Washington University,
  -- July, 1995.

-- Pragma Resource is a Janus pragma to build .rc files.
Pragma Resource(ICON,"Feldman", "Feldman.Ico");
Pragma Resource(ICON,"Barnes",  "Barnes.Ico");
Pragma Resource(ICON,"Cohen",   "NHC.Ico");
Pragma Resource(ICON,"Dewar",   "Dewar.Ico");
Pragma Resource(ICON,"Taft",    "Taft.Ico");

  -- philosophers sign into dining room, giving Maitre_D their DNA code

  type Philosopher_Ptr is access all Phil.Philosopher;

  A : aliased Phil.Philosopher (My_ID => 1);  -- Note: My_ID assumed to
  B : aliased Phil.Philosopher (My_ID => 2);  -- be in Table_Type and
  C : aliased Phil.Philosopher (My_ID => 3);  -- unique.
  D : aliased Phil.Philosopher (My_ID => 4);
  E : aliased Phil.Philosopher (My_ID => 5);

  Phils : array (Table_Type) of Philosopher_Ptr
    :=(A'Access,  -- order is not important
       C'Access,
       D'Access,
       B'Access,
       E'Access);

  Image_Icon : array(Table_Type) of Claw.Icons.Icon_Type;

  function Next_Stick(i: in Table_Type) return Table_Type is
  begin
    if i = Table_Type'last then return Table_Type'first;
    else return i+1;end if;
  end Next_Stick;

  task body Maitre_D is
    use type Phil.States;

    Display : Displays;

    task Seater is  -- If all diners are seated initially, the animation
      entry Start;  -- doesn't look as good, and deadlock may occur
    end Seater;     -- immediately.  So let them arrive over time.

    task body Seater is
    begin
      accept Start;
      for i in Table_Type loop
        delay 2.0;
        Phils(i).Start_Eating(Chopstick1 => Table_Type(Phils(i).My_ID),
                              Chopstick2 => Next_Stick(Table_Type(Phils(i).My_ID)));
      end loop;
    end Seater;

    Death_Toll : Natural := 0;

    Status : Status_Report_Type;

    Background_Brush : Claw.Brushes.Brush_Type
      := Claw.Brushes.Create(Claw.Brushes.Black);

    procedure Clean_Up is
    begin
      abort Seater;
      for i in Phils'range loop
        abort Phils(i).all;
      end loop;
    end Clean_Up;

  begin
    -- The Production version of Claw will normally shut down a
    -- "zombie" program whose main program has terminated.  We
    -- don't want that behavior here.  The simple Demo version never
    -- does that, and does not contain Set_Termination_Mode code.
    -- Claw.Set_Termination_Mode(Claw.Conservative);
    Claw.Icons.Create_Large(Image_Icon(1), "Feldman");
    Claw.Icons.Create_Large(Image_Icon(2), "Barnes");
    Claw.Icons.Create_Large(Image_Icon(3), "Cohen");
    Claw.Icons.Create_Large(Image_Icon(4), "Dewar");
    Claw.Icons.Create_Large(Image_Icon(5), "Taft");

    accept Start_Serving;

    Create(Window => Display,
           Window_Name => "Dining Philosophers");
    Show(Window => Display, How => Claw.Codes.Show_Startup);
    Set_Background_Brush(Display, Background_Brush);

    -- now Maitre_D assigns phils to seats at the table

    -- assign them their seats and chopsticks.

    Seater.Start;
    loop
      select
        accept Report_State (Which_Phil : in Society.Unique_DNA_Codes;
                             State      : in Phil.States;
                             How_Long   : in Natural := 0;
                             Which_Meal : in Natural := 0) do
          if State = Phil.Dying then
            Status(Which_Phil).Seated := False;
            Death_Toll := Death_Toll+1;
          else
            Status(Which_Phil).Seated := True;
          end if;
          Status(Which_Phil).Activity := State;
        end Report_State;
        Claw.Window_Operations.Invalidate(Display);
      or
        accept Status_Request(Snapshot : out Status_Report_Type) do
          Snapshot := Status;
        end Status_Request;
      or
        delay 0.1;  -- This was originally a 'terminate', but we
                    -- don't want to hang if the phils deadlock.
      end select;
      exit when Death_Toll >= Phils'length               -- all Phils done
      or not Is_Valid(Display);                          -- user closed window
    end loop;
    -- The user may have closed us down, don't leave any wounded.
    Clean_Up;
  exception
    when Claw.Not_Valid_Error =>  -- user slammed the window on us?
      Clean_Up;
    when oops:others =>
      Claw.Debug_Exception_Message_Box(oops, " Maitre_D killed");
      Clean_Up;
  end Maitre_D;

  procedure When_Draw(Window  : in out Displays;
                      Easel   : in out Claw.Canvas.Basic_Canvas_Type'Class;
                      Erase_Background: in Boolean;
                      Area    : in     Claw.Rectangle_Type) is
  -- This procedure is called whenever a section of the status
  -- window needs to be (re)drawn.  Draw is an open drawing object
  -- for the appropriate drawing.  Erase_Background is true if the
  -- background needs to be erased.  Area is the area to redraw.
    use type Claw.Int;

    Table_Rectangle : constant Claw.Rectangle_Type
      :=(Left=>64, Top=>64, Right=> 200, Bottom => 200);

    Table_Center : constant Claw.Point_Type
      := (X => (Table_Rectangle.Left + Table_Rectangle.Right)/2,
          Y => (Table_Rectangle.Top + Table_Rectangle.Bottom)/2);
    Table_Radius : constant Float
      := Float(Table_Rectangle.Right - Table_Rectangle.Left)/2.0;
    Chopstick_Offset_Angle : constant Float := 0.5/Float(Table_Size);
    Interior_Brush : Claw.Brushes.Brush_Type
      := Claw.Brushes.Create(Claw.Colors.RGB(Red=>200, Green=> 175,
                                             Blue=>150));
    White_Brush : Claw.Brushes.Brush_Type
      := Claw.Brushes.Create(Claw.Brushes.White);
    Thin_Pen : Claw.Pens.Pen_Type
      := Claw.Pens.Create(Claw.Pens.Black);
    Chopstick_Pen : Claw.Pens.Pen_Type
      := Claw.Pens.Create(Claw.Pens.Styles.Solid, Line_Width=>2,
           Color => Claw.Colors.RGB(0,0,0));

    Snapshot : Status_Report_Type;

    Chopstick_In_Use : array(Table_Type) of Boolean
      := (others => False);

    function XY(Radius : in Float; Fraction : in Float) return Claw.Point_Type is
      use Ada.Numerics.Elementary_Functions;
    begin
      return (X=>Table_Center.X+Claw.Int(Radius*Cos(Fraction*2.0*Ada.Numerics.Pi)),
              Y=>Table_Center.Y-Claw.Int(Radius*Sin(Fraction*2.0*Ada.Numerics.Pi)));
    end XY;

    function Phil_Angle(i : in Table_Type) return Float is
    begin return Float(i)/Float(Table_Size);end Phil_Angle;

    function Picture_Top_Left(i : in Table_Type) return Claw.Point_Type is
      -- Place picture outside, but close to, the table.
      Picture_Size : constant Claw.Int := 32; -- square
      Picture_Radius : constant Claw.Int := Claw.Int(0.707*Float(Picture_Size));
      Center : constant Claw.Point_Type
        := XY(Table_Radius+Float(Picture_Radius), Phil_Angle(i));
    begin
      return (X=>Center.X-Picture_Size/2, Y=>Center.Y-Picture_Size/2);
    end Picture_Top_Left;

    function Plate_Rectangle(i : in Table_Type) return Claw.Rectangle_Type is
      Center : constant Claw.Point_Type
        := XY(0.75*Table_Radius, Phil_Angle(i));
    begin
      return (Top=>Center.Y-10, Left=>Center.X-10,
              Bottom=>Center.Y+10, Right=>Center.X+10);
    end Plate_Rectangle;

    function In_Use_Chopstick_Tip(Phil : in Table_Type) return Claw.Point_Type is
    begin
      return XY(Table_Radius*0.65, Phil_Angle(Phil));
    end In_Use_Chopstick_Tip;

    function Chopstick_Base(Stick : in Table_Type) return Claw.Point_Type is
    begin
      return XY(Table_Radius*0.95, Phil_Angle(Stick)-Chopstick_Offset_Angle);
    end Chopstick_Base;

    function Avail_Chopstick_Tip(Stick : in Table_Type) return Claw.Point_Type is
    begin
      return XY(Table_Radius*0.50, Phil_Angle(Stick)-Chopstick_Offset_Angle);
    end Avail_Chopstick_Tip;

  begin
    select
      Maitre_D.Status_Request(Snapshot);
    or            -- never hang indefinitely in a When_xxx routine!
      delay 0.1;
      return; -- no info => no picture (Maitre_D busy opening/closing?)
    end select;

    Claw.Canvas.Choose(Easel, Interior_Brush);
    Claw.Canvas.Choose(Easel, Thin_Pen);
    Claw.Draw.Ellipse(Easel, Table_Rectangle);

    Claw.Canvas.Choose(Easel, White_Brush);
    for i in Table_Type loop
      Claw.Draw.Ellipse(Easel, Plate_Rectangle(i));
    end loop;

    Claw.Canvas.Choose(Easel, Chopstick_Pen);

    for i in Table_Type'range loop
      if Snapshot(i).Seated then
        Claw.Icons.Draw (Image_Icon(i), Easel, Picture_Top_Left(i));
      end if;
      case Snapshot(i).Activity is
        when Phil.Got_One_Stick =>
          Chopstick_In_Use(i) := True;
          Claw.Draw.Set_Current_Position(Easel, Chopstick_Base(i));
          Claw.Draw.Line_To(Easel, In_Use_Chopstick_Tip(i));
        when Phil.Got_Other_Stick | Phil.Eating | Phil.Done_Eating =>
          Chopstick_In_Use(i) := True;
          Claw.Draw.Set_Current_Position(Easel, Chopstick_Base(i));
          Claw.Draw.Line_To(Easel, In_Use_Chopstick_Tip(i));
          Chopstick_In_Use(Next_Stick(i)) := True;
          Claw.Draw.Line_To(Easel, Chopstick_Base(Next_Stick(i)));
        when others => null;
      end case;
    end loop;

    for i in Chopstick_In_Use'range loop
      if not Chopstick_In_Use(i) then
        Claw.Draw.Set_Current_Position(Easel, Chopstick_Base(i));
        Claw.Draw.Line_To(Easel, Avail_Chopstick_Tip(i));
      end if;
    end loop;
  exception
    when others => null;  -- can't draw, so don't.
                          -- But don't kill calling message_loop task
  end When_Draw;

end Room;
